package ru.landgrafhomyak.db.serdha0.api.module

import ru.landgrafhomyak.db.serdha0.api.table.Table
import ru.landgrafhomyak.db.serdha0.api.queries.QueryCreator
import ru.landgrafhomyak.db.serdha0.api.runtime.SynchronizedDatabase
import ru.landgrafhomyak.db.serdha0.api.runtime.Transaction
import ru.landgrafhomyak.db.serdha0.api.table.CreateTableScope

/**
 * Scope methods for creating new [module template][ModuleTemplate].
 *
 * @see ModuleTemplate.CreateModuleTemplatesScope.createTemplate
 * @see ModuleTemplate.ModuleCreator.createSchema
 */
public interface CreateModuleScope {
	/**
	 * Functional interface for creating table.
	 *
	 * Can be implemented like a companion object that just calls [TableUserExtension]'s constructor with the same or similar signature.
	 *
	 * @param TableUserExtension User's type for containing table-related descriptors.
	 *
	 * @see CreateModuleScope.createTable
	 * @see CreateModuleScope.createSessionScopeTemporaryTable
	 * @see CreateModuleScope.createTransactionScopeTemporaryTable
	 * @see UpgradeModuleScope.TableUpgrade
	 */
	public interface TableCreator<TableUserExtension : Any> {
		/**
		 * Scope method that uses [creator] to initialize table schema.
		 *
		 * @param creator Object with descriptor providers and factories.
		 * @return User's object with descriptors for future access.
		 *
		 * @see CreateModuleScope.createTable
		 * @see CreateModuleScope.createSessionScopeTemporaryTable
		 * @see CreateModuleScope.createTransactionScopeTemporaryTable
		 * @see UpgradeModuleScope.TableUpgrade
		 */
		public fun createTable(creator: CreateTableScope<TableUserExtension>): TableUserExtension
	}

	/**
	 * Creates table in [specified namespace][namespace] with [specified name][name] and stores it in the database.
	 * If table was created by previous synchronizations, asserts that table at the specified path in the database is same as table provided by [initializer].
	 *
	 * On [module upgrading][ModuleTemplate.CreateModuleTemplatesScope.upgradeTemplate] this table must be [kept][UpgradeModuleScope.upgradeTable] (or [renamed][UpgradeModuleScope.renameTable]) or [deleted][UpgradeModuleScope.deleteTable].
	 *
	 * @param TableUserExtension User's type for containing table-related descriptors.
	 * @param namespace Way to group tables if there is a lot.
	 * @param name Name of table for debugging.
	 * @param initializer Table-related descriptors initializer.
	 * @return Table descriptor.
	 */
	public fun <TableUserExtension : Any> createTable(namespace: Namespace, name: String, initializer: TableCreator<TableUserExtension>): Table<TableUserExtension, Nothing>

	/**
	 * Creates a temporary table that exits only until connection to the database not [closed][SynchronizedDatabase._close].
	 *
	 * On [module upgrading][ModuleTemplate.CreateModuleTemplatesScope.upgradeTemplate] must be recreated if needed.
	 *
	 * @param TableUserExtension User's type for containing table-related descriptors.
	 * @param namespace Way to group tables if there is a lot.
	 * @param name Name of table for debugging.
	 * @param initializer Table-related descriptors initializer.
	 * @return Table descriptor.
	 */
	public fun <TableUserExtension : Any> createSessionScopeTemporaryTable(namespace: Namespace, name: String, initializer: TableCreator<TableUserExtension>): Table<TableUserExtension, Nothing>

	/**
	 * Creates a temporary table that exits only inside [transaction][Transaction] and auto-deleted (or cleared, depends on driver implementation) when it [finished][Transaction._assertTransactionFinishedAndReleaseResources].
	 *
	 * On [module upgrading][ModuleTemplate.CreateModuleTemplatesScope.upgradeTemplate] must be recreated if needed.
	 *
	 * @param TableUserExtension User's type for containing table-related descriptors.
	 * @param namespace Way to group tables if there is a lot.
	 * @param name Name of table for debugging.
	 * @param initializer Table-related descriptors initializer.
	 * @return Table descriptor.
	 */
	public fun <TableUserExtension : Any> createTransactionScopeTemporaryTable(namespace: Namespace, name: String, initializer: TableCreator<TableUserExtension>): Table<TableUserExtension, Nothing>

	public val queries: QueryCreator.Scope

	public val temporarySubqueries: QueryCreator.Scope

	public val moduleConfiguringTemporaryQueries: QueryCreator.Scope

	public fun <qUE : Any> createQuery(creator: QueryCreator<qUE>): qUE

	public fun <qUE : Any> createModuleConfiguringTemporaryQuery(creator: QueryCreator<qUE>): qUE

	/**
	 * Replaces [specified namespace][rootNs] with schema provided by [template].
	 * [This namespace][rootNs] will be root namespace for module and shouldn't contain any other definitions.
	 *
	 * @param rootNs Namespace where the new module will be located.
	 * @param template Schema template.
	 * @return Module descriptor.
	 */
	public fun <ModuleUserExtension : Any> substituteModule(rootNs: Namespace, template: ModuleTemplate<ModuleUserExtension>): Module<ModuleUserExtension>
}